@@ -39,6 +39,7 @@ module Text.Hamlet
     , hamletRules
     , ihamletRules
     , htmlRules
+    , CloseStyle (..)
     ) where
 
 import Text.Shakespeare.Base
@@ -55,6 +56,7 @@ import qualified Data.Foldable as F
 import Control.Monad (mplus)
 import Data.Monoid (mempty, mappend)
 import Control.Arrow ((***))
+import Data.List (intercalate)
 
 -- | Convert some value to a list of attribute pairs.
 class ToAttributes a where
@@ -101,15 +103,72 @@ unIdent :: Ident -> String
 unIdent (Ident s) = s
 
 bindingPattern :: Binding -> Q (Pat, [(Ident, Exp)])
-bindingPattern (BindVar i@(Ident s)) = do
+bindingPattern (BindAs i@(Ident s) b) = do
     name <- newName s
-    return (VarP name, [(i, VarE name)])
+    (pattern, scope) <- bindingPattern b
+    return (AsP name pattern, (i, VarE name):scope)
+bindingPattern (BindVar i@(Ident s))
+    | all isDigit s = do
+        return (LitP $ IntegerL $ read s, [])
+    | otherwise = do
+        name <- newName s
+        return (VarP name, [(i, VarE name)])
 bindingPattern (BindTuple is) = do
-    names <- mapM (newName . unIdent) is
-    return (TupP $ map VarP names, zip is $ map VarE names)
-bindingPattern (BindConstr (Ident con) is) = do
-    names <- mapM (newName . unIdent) is
-    return (ConP (mkName con) (map VarP names), zip is $ map VarE names)
+    (patterns, scopes) <- fmap unzip $ mapM bindingPattern is
+    return (TupP patterns, concat scopes)
+bindingPattern (BindList is) = do
+    (patterns, scopes) <- fmap unzip $ mapM bindingPattern is
+    return (ListP patterns, concat scopes)
+bindingPattern (BindConstr con is) = do
+    (patterns, scopes) <- fmap unzip $ mapM bindingPattern is
+    return (ConP (mkConName con) patterns, concat scopes)
+bindingPattern (BindRecord con fields wild) = do
+    let f (Ident field,b) =
+           do (p,s) <- bindingPattern b
+              return ((mkName field,p),s)
+    (patterns, scopes) <- fmap unzip $ mapM f fields
+    (patterns1, scopes1) <- if wild
+       then bindWildFields con $ map fst fields
+       else return ([],[])
+    return (RecP (mkConName con) (patterns++patterns1), concat scopes ++ scopes1)
+
+mkConName :: DataConstr -> Name
+mkConName = mkName . conToStr
+
+conToStr :: DataConstr -> String
+conToStr (DCUnqualified (Ident x)) = x
+conToStr (DCQualified (Module xs) (Ident x)) = intercalate "." $ xs ++ [x]
+
+-- Wildcards bind all of the unbound fields to variables whose name
+-- matches the field name.
+--
+-- For example: data R = C { f1, f2 :: Int }
+-- C {..}           is equivalent to   C {f1=f1, f2=f2}
+-- C {f1 = a, ..}   is equivalent to   C {f1=a,  f2=f2}
+-- C {f2 = a, ..}   is equivalent to   C {f1=f1, f2=a}
+bindWildFields :: DataConstr -> [Ident] -> Q ([(Name, Pat)], [(Ident, Exp)])
+bindWildFields conName fields = do
+  fieldNames <- recordToFieldNames conName
+  let available n     = nameBase n `notElem` map unIdent fields
+  let remainingFields = filter available fieldNames
+  let mkPat n = do
+        e <- newName (nameBase n)
+        return ((n,VarP e), (Ident (nameBase n), VarE e))
+  fmap unzip $ mapM mkPat remainingFields
+
+-- Important note! reify will fail if the record type is defined in the
+-- same module as the reify is used. This means quasi-quoted Hamlet
+-- literals will not be able to use wildcards to match record types
+-- defined in the same module.
+recordToFieldNames :: DataConstr -> Q [Name]
+recordToFieldNames conStr = do
+  -- use 'lookupValueName' instead of just using 'mkName' so we reify the
+  -- data constructor and not the type constructor if their names match.
+  Just conName                <- lookupValueName $ conToStr conStr
+  DataConI _ _ typeName _     <- reify conName
+  TyConI (DataD _ _ _ cons _) <- reify typeName
+  [fields] <- return [fields | RecC name fields <- cons, name == conName]
+  return [fieldName | (fieldName, _, _) <- fields]
 
 docToExp :: Env -> HamletRules -> Scope -> Doc -> Q Exp
 docToExp env hr scope (DocForall list idents inside) = do
@@ -157,7 +216,7 @@ docToExp env hr scope (DocCond conds final) = do
   where
     go :: (Deref, [Doc]) -> Q Exp
     go (d, docs) = do
-        let d' = derefToExp scope d
+        let d' = derefToExp ((specialOrIdent, VarE 'or):scope) d
         docs' <- docsToExp env hr scope docs
         return $ TupE [d', docs']
 docToExp env hr scope (DocCase deref cases) = do
@@ -169,14 +228,11 @@ docToExp env hr scope (DocCase deref cases) = do
         case reads s of
             (x, ""):_ -> Just x
             _ -> Nothing
+    toMatch :: (Binding, [Doc]) -> Q Match
     toMatch (idents, inside) = do
-        let pat = case map unIdent idents of
-                    ["_"] -> WildP
-                    [str]
-                        | Just i <- readMay str -> LitP $ IntegerL i
-                    strs -> let (constr:fields) = map mkName strs
-                            in ConP constr (map VarP fields)
-        insideExp <- docsToExp env hr scope inside
+        (pat, extraScope) <- bindingPattern idents
+        let scope' = extraScope ++ scope
+        insideExp <- docsToExp env hr scope' inside
         return $ Match pat (NormalB insideExp) []
 docToExp env hr v (DocContent c) = contentToExp env hr v c
 
@@ -290,7 +346,7 @@ hamletFromString qhr set s = do
     hr <- qhr
     case parseDoc set s of
         Error s' -> error s'
-        Ok (mnl, d) -> hrWithEnv hr $ \env -> docsToExp env hr [] d
+        Ok (_mnl, d) -> hrWithEnv hr $ \env -> docsToExp env hr [] d
 
 hamletFileWithSettings :: Q HamletRules -> HamletSettings -> FilePath -> Q Exp
 hamletFileWithSettings qhr set fp = do
