@@ -1,4 +1,5 @@
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE CPP #-}
 module Text.Hamlet.Parse
     ( Result (..)
@@ -11,6 +12,9 @@ module Text.Hamlet.Parse
     , CloseStyle (..)
     , Binding (..)
     , NewlineStyle (..)
+    , specialOrIdent
+    , DataConstr (..)
+    , Module (..)
     )
     where
 
@@ -18,11 +22,12 @@ import Text.Shakespeare.Base
 import Control.Applicative ((<$>), Applicative (..))
 import Control.Monad
 import Control.Arrow
+import Data.Char (isUpper)
 import Data.Data
 import Text.ParserCombinators.Parsec hiding (Line)
 import Data.Set (Set)
 import qualified Data.Set as Set
-import Data.Maybe (mapMaybe, fromMaybe)
+import Data.Maybe (mapMaybe, fromMaybe, isNothing)
 
 data Result v = Error String | Ok v
     deriving (Show, Eq, Read, Data, Typeable)
@@ -53,7 +58,7 @@ data Line = LineForall Deref Binding
           | LineMaybe Deref Binding
           | LineNothing
           | LineCase Deref
-          | LineOf [Ident]
+          | LineOf Binding
           | LineTag
             { _lineTagName :: String
             , _lineAttr :: [(Maybe Deref, String, Maybe [Content])]
@@ -100,6 +105,7 @@ parseLine set = do
     x <- doctype <|>
          doctypeDollar <|>
          comment <|>
+         ssiInclude <|>
          htmlComment <|>
          doctypeRaw <|>
          backslash <|>
@@ -150,6 +156,11 @@ parseLine set = do
         _ <- many $ noneOf "\r\n"
         eol
         return $ LineContent [] True
+    ssiInclude = do
+        x <- try $ string "<!--#"
+        y <- many $ noneOf "\r\n"
+        eol
+        return $ LineContent [ContentRaw $ x ++ y] False
     htmlComment = do
         _ <- try $ string "<!--"
         _ <- manyTill anyChar $ try $ string "-->"
@@ -215,10 +226,11 @@ parseLine set = do
         return $ LineCase x
     controlOf = do
         _   <- try $ string "$of"
-        pat <- many1 $ try $ spaces >> ident
+        spaces
+        x <- identPattern
         _   <- spaceTabs
         eol
-        return $ LineOf pat
+        return $ LineOf x
     content cr = do
         x <- many $ content' cr
         case cr of
@@ -266,7 +278,13 @@ parseLine set = do
     tagCond = do
         d <- between (char ':') (char ':') parseDeref
         tagClass (Just d) <|> tagAttrib (Just d)
-    tagClass x = char '.' >> (TagClass . (,) x) <$> tagAttribValue NotInQuotes
+    tagClass x = do
+        clazz <- char '.' >> tagAttribValue NotInQuotes
+        let hasHash (ContentRaw s) = any (== '#') s
+            hasHash _ = False
+        if any hasHash clazz
+            then fail $ "Invalid class: " ++ show clazz ++ ". Did you want a space between a class and an ID?"
+            else return (TagClass (x, clazz))
     tagAttrib cond = do
         s <- many1 $ noneOf " \t=\r\n><"
         v <- (char '=' >> Just <$> tagAttribValue NotInQuotesAttr) <|> return Nothing
@@ -285,20 +303,109 @@ parseLine set = do
     tag'' (TagAttribs s) (x, y, z, as) = (x, y, z, s : as)
 
     ident :: Parser Ident
-    ident = Ident <$> many1 (alphaNum <|> char '_' <|> char '\'')
+    ident = do
+      i <- many1 (alphaNum <|> char '_' <|> char '\'')
+      white
+      return (Ident i)
+     <?> "identifier"
+
+    parens = between (char '(' >> white) (char ')' >> white)
+
+    brackets = between (char '[' >> white) (char ']' >> white)
+
+    braces = between (char '{' >> white) (char '}' >> white)
+
+    comma = char ',' >> white
+
+    atsign = char '@' >> white
+
+    equals = char '=' >> white
+
+    white = skipMany $ char ' '
+
+    wildDots = string ".." >> white
+
+    isVariable (Ident (x:_)) = not (isUpper x)
+    isVariable (Ident []) = error "isVariable: bad identifier"
+
+    isConstructor (Ident (x:_)) = isUpper x
+    isConstructor (Ident []) = error "isConstructor: bad identifier"
 
     identPattern :: Parser Binding
-    identPattern = (between
-        (char '(' >> spaces)
-        (spaces >> char ')' >> spaces)
-        (BindTuple <$> sepBy1 ident (spaces >> char ',' >> spaces))
-        ) <|> (do
-            i <- ident
-            is <- many $ try $ (many $ char ' ') >> ident
-            if null is
-                then return $ BindVar i
-                else return $ BindConstr i is
-            )
+    identPattern = gcon True <|> apat
+      where
+      apat = choice
+        [ varpat
+        , gcon False
+        , parens tuplepat
+        , brackets listpat
+        ]
+
+      varpat = do
+        v <- try $ do v <- ident
+                      guard (isVariable v)
+                      return v
+        option (BindVar v) $ do
+          atsign
+          b <- apat
+          return (BindAs v b)
+       <?> "variable"
+
+      gcon :: Bool -> Parser Binding
+      gcon allowArgs = do
+        c <- try $ do c <- dataConstr
+                      return c
+        choice
+          [ record c
+          , fmap (BindConstr c) (guard allowArgs >> many apat)
+          , return (BindConstr c [])
+          ]
+       <?> "constructor"
+
+      dataConstr = do
+        p <- dcPiece
+        ps <- many dcPieces
+        return $ toDataConstr p ps
+
+      dcPiece = do
+        x@(Ident y) <- ident
+        guard $ isConstructor x
+        return y
+
+      dcPieces = do
+        _ <- char '.'
+        dcPiece
+
+      toDataConstr x [] = DCUnqualified $ Ident x
+      toDataConstr x (y:ys) =
+          go (x:) y ys
+        where
+          go front next [] = DCQualified (Module $ front []) (Ident next)
+          go front next (rest:rests) = go (front . (next:)) rest rests
+
+      record c = braces $ do
+        (fields, wild) <- option ([], False) $ go
+        return (BindRecord c fields wild)
+        where
+        go = (wildDots >> return ([], True))
+           <|> (do x         <- recordField
+                   (xs,wild) <- option ([],False) (comma >> go)
+                   return (x:xs,wild))
+
+      recordField = do
+        field <- ident
+        p <- option (BindVar field) -- support punning
+                    (equals >> identPattern)
+        return (field,p)
+
+      tuplepat = do
+        xs <- identPattern `sepBy` comma
+        return $ case xs of
+          [x] -> x
+          _   -> BindTuple xs
+
+      listpat = BindList <$> identPattern `sepBy` comma
+
     angle = do
         _ <- char '<'
         name' <- many  $ noneOf " \t.#\r\n!>"
@@ -334,7 +441,7 @@ data Doc = DocForall Deref Binding [Doc]
          | DocWith [(Deref, Binding)] [Doc]
          | DocCond [(Deref, [Doc])] (Maybe [Doc])
          | DocMaybe Deref Binding [Doc] (Maybe [Doc])
-         | DocCase Deref [([Ident], [Doc])]
+         | DocCase Deref [(Binding, [Doc])]
          | DocContent Content
     deriving (Show, Eq, Read, Data, Typeable)
 
@@ -364,9 +471,9 @@ nestToDoc set (Nest (LineMaybe d i) inside:rest) = do
     rest'' <- nestToDoc set rest'
     Ok $ DocMaybe d i inside' nothing : rest''
 nestToDoc set (Nest (LineCase d) inside:rest) = do
-    let getOf (Nest (LineOf pat) insideC) = do
+    let getOf (Nest (LineOf x) insideC) = do
             insideC' <- nestToDoc set insideC
-            Ok (pat, insideC')
+            Ok (x, insideC')
         getOf _ = Error "Inside a $case there may only be $of.  Use '$of _' for a wildcard."
     cases <- mapM getOf inside
     rest' <- nestToDoc set rest
@@ -381,7 +488,7 @@ nestToDoc set (Nest (LineTag tn attrs content classes attrsD avoidNewLine) insid
     let attrs' =
             case clazzes of
               [] -> map attrFix noclass
-              _ -> (Nothing, "class", map (second Just) clazzes)
+              _ -> (testIncludeClazzes clazzes, "class", map (second Just) clazzes)
                        : map attrFix noclass
     let closeStyle =
             if not (null content) || not (null inside)
@@ -570,8 +677,41 @@ doctypeNames =
     , ("strict", "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">")
     ]
 
-data Binding = BindVar Ident | BindConstr Ident [Ident] | BindTuple [Ident]
+data Binding = BindVar Ident
+             | BindAs Ident Binding
+             | BindConstr DataConstr [Binding]
+             | BindTuple [Binding]
+             | BindList [Binding]
+             | BindRecord DataConstr [(Ident, Binding)] Bool
+    deriving (Eq, Show, Read, Data, Typeable)
+
+data DataConstr = DCQualified Module Ident
+                | DCUnqualified Ident
+    deriving (Eq, Show, Read, Data, Typeable)
+
+newtype Module = Module [String]
     deriving (Eq, Show, Read, Data, Typeable)
 
 spaceTabs :: Parser String
 spaceTabs = many $ oneOf " \t"
+
+-- | When using conditional classes, it will often be a single class, e.g.:
+--
+-- > <div :isHome:.homepage>
+--
+-- If isHome is False, we do not want any class attribute to be present.
+-- However, due to combining multiple classes together, the most obvious
+-- implementation would produce a class="". The purpose of this function is to
+-- work around that. It does so by checking if all the classes on this tag are
+-- optional. If so, it will only include the class attribute if at least one
+-- conditional is true.
+testIncludeClazzes :: [(Maybe Deref, [Content])] -> Maybe Deref
+testIncludeClazzes cs
+    | any (isNothing . fst) cs = Nothing
+    | otherwise = Just $ DerefBranch (DerefIdent specialOrIdent) $ DerefList $ mapMaybe fst cs
+
+-- | This funny hack is to allow us to refer to the 'or' function without
+-- requiring the user to have it in scope. See how this function is used in
+-- Text.Hamlet.
+specialOrIdent :: Ident
+specialOrIdent = Ident "__or__hamlet__special"
