@@ -66,11 +66,15 @@ parseHamletRT set s =
         Error s' -> failure $ HamletParseException s'
         Ok (_, x) -> liftM HamletRT $ mapM convert x
   where
+    convert x@(DocForall deref (BindAs _ _) docs) =
+       error "Runtime Hamlet does not currently support 'as' patterns"
     convert x@(DocForall deref (BindVar (Ident ident)) docs) = do
         deref' <- flattenDeref' x deref
         docs' <- mapM convert docs
         return $ SDForall deref' ident docs'
     convert DocForall{} = error "Runtime Hamlet does not currently support tuple patterns"
+    convert x@(DocMaybe deref (BindAs _ _) jdocs ndocs) =
+       error "Runtime Hamlet does not currently support 'as' patterns"
     convert x@(DocMaybe deref (BindVar (Ident ident)) jdocs ndocs) = do
         deref' <- flattenDeref' x deref
         jdocs' <- mapM convert jdocs
@@ -97,6 +101,11 @@ parseHamletRT set s =
         els' <- maybe (return []) (mapM convert) els
         return $ SDCond conds' els'
       where
+        -- | See the comments in Text.Hamlet.Parse.testIncludeClazzes. The conditional
+        -- added there doesn't work for runtime Hamlet, so we remove it here.
+        go (DerefBranch (DerefIdent x) _, docs') | x == specialOrIdent = do
+            docs'' <- mapM convert docs'
+            return (["True"], docs'')
         go (deref, docs') = do
             deref' <- flattenDeref' x deref
             docs'' <- mapM convert docs'
@@ -172,6 +181,7 @@ renderHamletRT' tempAsHtml (HamletRT docs) scope0 renderUrl =
             => [String] -> [String] -> HamletMap url -> m (HamletData url)
     lookup' orig k m =
         case lookup k m of
+            Nothing | k == ["True"] -> return $ HDBool True
             Nothing -> fa $ showName orig ++ ": not found"
             Just x -> return x
 
